from .wordvec import extract_embeddings
from .wordvec import load_word2vec
from .wordvec import load_fasttext
from .wordvec import build_vocab_from_word2vec


def is_chinese(char):
    """
    判断是否汉字

    汉字 Unicode 编码范围 (http://www.qqxiuzi.cn/zh/hanzi-unicode-bianma.php)
        字符集	    字数	Unicode 编码
        基本汉字	    20902字	    4E00-9FA5
        基本汉字补充	38字	        9FA6-9FCB
        扩展A	    6582字	    3400-4DB5
        扩展B	    42711字	    20000-2A6D6
        扩展C	    4149字	    2A700-2B734
        扩展D	    222字	    2B740-2B81D
        康熙部首	    214字	    2F00-2FD5
        部首扩展	    115字	    2E80-2EF3
        兼容汉字	    477字	    F900-FAD9
        兼容扩展	    542字	    2F800-2FA1D
        PUA(GBK)部件	81字	        E815-E86F
        部件扩展	    452字	    E400-E5E8
        PUA增补	    207字	    E600-E6CF
        汉字笔画	    36字	        31C0-31E3
        汉字结构	    12字	        2FF0-2FFB
        汉语注音	    22字	        3105-3120
        注音扩展	    22字	        31A0-31BA
        〇	        1字	        3007

    Examples:
        >>> is_chinese("华")
        True
        >>> is_chinese("a")
        False

    Args:
        char:

    Returns:

    """
    if '\u4e00' <= char <= '\u9fa5':
        return True
    return False


def _remove_duplicate(src, dst=None, encoding="utf8"):
    """
    Examples:
        >>> _remove_duplicate("data/stopwords_zh")
    """
    with open(src, encoding=encoding) as f:
        s = set()
        l = list()
        for i in f:
            i = i.strip().lower()
            if i not in s:
                s.add(i)
                l.append(i)

    if dst is None:
        dst = src

    with open(dst, 'w', encoding=encoding) as f:
        for i in l:
            f.write(i)
            f.write('\n')


def load_stopwords(filepath, encoding="utf8"):
    """
    Args:
        filepath(str):
        encoding:

    Examples:
        >>> stopwords_en = load_stopwords("./data/stopwords_en")
        >>> 'the' in stopwords_en
        True

    Returns:
        set
    """
    with open(filepath, encoding=encoding) as f:
        stopwords = set(word.strip().lower() for word in f if not word.isspace())

    return stopwords


def remove_stopwords(tokens, stopwords=None, encoding='utf8'):
    """remove stopwords

    Args:
        tokens(list of str): a list of tokens/words
        stopwords(list of str or str): a list of words or a filepath
        encoding:

    Examples:
        >>> from huaytools.nlp.stopwords import stopwords_en
        >>> remove_stopwords(['huay', 'the'], stopwords_en)
        ['huay']

    """
    if stopwords is None:
        return tokens

    if isinstance(stopwords, str):
        stopwords = load_stopwords(stopwords, encoding)
    else:
        stopwords = set(stopwords)

    return [w for w in tokens if w not in stopwords]


def remain_chinese(sequences):
    """
    保留中文字符

    Todo(huay):
    """


def clear_str_en(text):
    """clear the en text

    Todo(huay):
    """



"""
Handle the different kinds of word embedding
    * word2vec
    * GloVe
    * FastText
"""
import os
import json
import pickle
import logging
import numpy as np

logger = logging.getLogger(__name__)


def load_word2vec(wv_model_path, binary=False):
    """
    使用 gensim 加载 Word2Vec 模型

    Args:
        wv_model_path(str): 词向量模型路径
        binary(bool): 词向量模型的类型；
            若为 True，则为 gensim 保存的二进制格式；反之为原始 C word2vec-tool 生成的格式

        原始 C word2vec-tool 生成的格式为：
            ```
            2519370 300
            the -0.065334 -0.093031 -0.017571 ...
            of 0.048804 -0.28528 0.018557 ...
            ```
    """
    logger.info("loading the Word2Vec model...")

    if not binary:
        from gensim.models import KeyedVectors
        wv_model = KeyedVectors.load_word2vec_format(wv_model_path)
    else:
        from gensim.models import Word2Vec
        wv_model = Word2Vec.load(wv_model_path)

    logger.info("loading the Word2Vec model finished.")
    return wv_model


def load_fasttext(model_path):
    """
    加载 FastText 模型（加载时间较长）

    Args:
        model_path(str): /path/to/model.bin
            fasttext GitHub 提供了预训练的模型，可以直接导入 gensim
    """
    logger.info("loading the FastText model...")

    from gensim.models import FastText
    model = FastText.load_fasttext_format(model_path)

    logger.info("loading the FastText model finished.")
    return model


def extract_embeddings(wv_model_path, words, save_path=None, binary=False):
    """
    从 wv_model 中抽取 words 的词向量

    Args:
        wv_model_path(str): 预训练的词向量模型路径
        words(list): 待抽取词向量的单词
        save_path(str): 文件保存路径
        binary(bool): 词向量模型的类型

    Examples:
        >>> wv_model_path = r""
        >>> words = ['the', 'of', 'apple', 'UNK']
        >>> embeddings, id2word, word2id = extract_embeddings(wv_model_path, words, save_path='D:/tmp')

    Returns:

    """
    wv_model = load_word2vec(wv_model_path, binary=binary)

    embeddings = []
    id2word = dict()
    for id, word in enumerate(words):
        try:
            embeddings.append(wv_model[word])
            id2word[id + 1] = word
        except KeyError:
            pass

    embeddings = np.array(embeddings)

    from collections import defaultdict
    word2id = defaultdict(int, zip(id2word.values(), id2word.keys()))

    if save_path is not None:
        np.save(os.path.join(save_path, "embeddings.npy"), embeddings)
        with open(os.path.join(save_path, "id2word.json"), 'w', encoding='utf8') as f:
            json.dump(id2word, f, indent=2)
        with open(os.path.join(save_path, "word2id.pkl"), 'wb') as f:
            pickle.dump(word2id, f)

    return embeddings, id2word, word2id


def build_vocab_from_word2vec(wv_model_path, save_path=None, save_embeddings=False, binary=False, zero_start=True):
    """
    从训练好的词向量模型中构建词汇集

    Args:
        wv_model_path(str): 预训练的词向量模型路径
        save_path(str): 文件保存路径
        save_embeddings(bool): 是否保存 embeddings，默认不保存
        binary(bool): 词向量模型的类型
        zero_start(bool): id2word 是否从 0 开始
            若从 0 开始，即不留位置给未登录词，未登录需要在传入模型前处理
            若从 1 开始，则 0 的位置留给未登录词，所有未登录词将被映射到 id == 0；
                此时，一般的做法是将一个 0 向量拼接到所有词向量之前，所有未登录词将映射到该 0 向量

    Returns:

    """
    wv_model = load_word2vec(wv_model_path, binary=binary)
    embeddings = wv_model.wv.vectors

    if zero_start:
        id2word = {i: j for i, j in enumerate(wv_model.wv.index2word)}
        word2id = dict(zip(id2word.values(), id2word.keys()))
    else:
        id2word = {i + 1: j for i, j in enumerate(wv_model.wv.index2word)}
        word2id = dict(zip(id2word.values(), id2word.keys()))
        from collections import defaultdict
        word2id = defaultdict(int, word2id)

    if save_path is not None:
        with open(os.path.join(save_path, "id2word.json"), 'w', encoding='utf8') as f:
            json.dump(id2word, f, indent=2)
        with open(os.path.join(save_path, "word2id.pkl"), 'wb') as f:
            pickle.dump(word2id, f)

        if save_embeddings:
            np.save(os.path.join(save_path, "embeddings.npy"), embeddings)

    return id2word, word2id, embeddings


def build_vocab_from_fasttext(model_path, save_path=None, save_embeddings=False, zero_start=True):
    """"""
    model = load_fasttext(model_path)
    embeddings = model.wv.vectors

    if zero_start:
        id2word = {i: j for i, j in enumerate(model.wv.index2word)}
        word2id = dict(zip(id2word.values(), id2word.keys()))
    else:
        id2word = {i + 1: j for i, j in enumerate(model.wv.index2word)}
        word2id = dict(zip(id2word.values(), id2word.keys()))
        from collections import defaultdict
        word2id = defaultdict(int, word2id)

    if save_path is not None:
        with open(os.path.join(save_path, "id2word.json"), 'w', encoding='utf8') as f:
            json.dump(id2word, f, indent=2)
        with open(os.path.join(save_path, "word2id.pkl"), 'wb') as f:
            pickle.dump(word2id, f)

        if save_embeddings:
            np.save(os.path.join(save_path, "embeddings.npy"), embeddings)

    return id2word, word2id, embeddings



import os
import doctest


def test_bunch():
    from huaytools.utils import bunch
    doctest.testmod(bunch)


def test_utils():
    from huaytools import utils
    doctest.testmod(utils)


def test_stopwords():
    from huaytools.nlp import stopwords
    sw_char = stopwords.stopwords_char

    from itertools import islice
    for i in islice(sw_char, 10):
        print(i)


if __name__ == '__main__':
    """"""
    # test_bunch()
    # test_utils()
    test_stopwords()


""""""
import os
import sys
import pickle
import logging

from six.moves import urllib

from .bunch import Bunch, bunchify, unbunchify
from .time import *


def maybe_mkdirs(path, is_file=False, exist_ok=True):
    """递归创建文件夹

    Args:
        path (str): 待创建的路径，递归创建
        is_file (bool): 是否为文件路径
        exist_ok (bool): 默认为 True

    Examples:

        >>> maybe_mkdirs('D:/Tmp/a/b/')
        'D:/Tmp/a/b/'
        >>> maybe_mkdirs('D:/Tmp/a/b/c.txt')
        'D:/Tmp/a/b/c.txt'
        >>> maybe_mkdirs('D:/Tmp/a/b/c', is_file=True)  # 假设 c 是一个无后缀文件
        'D:/Tmp/a/b/c'

    Returns:
        str
    """
    if is_file:
        dirs, filename = os.path.split(path)
        os.makedirs(dirs, exist_ok=exist_ok)
    else:
        os.makedirs(path, exist_ok=exist_ok)

    return path


def save_to_pickle(obj, filepath):
    """
    保存到 pickle 文件

    Args:
        obj: 需要保存的对象
        filepath(str): 文件名

    Returns:
        None

    """
    filepath = maybe_mkdirs(filepath, is_file=True)
    with open(filepath, 'wb') as f:
        pickle.dump(obj, f)


def load_from_pickle(filepath):
    """
    从 pickle 加载对象

    Args:
        filepath(str): 文件名

    Returns:

    """
    with open(filepath) as f:
        return pickle.load(f)


def set_logging_basic_config(**kwargs):
    """
    快速设置 logging.basicConfig

    Args can be specified:
        filename: Specifies that a FileHandler be created, using the specified
            filename, rather than a StreamHandler.
        filemode: Specifies the mode to open the file, if filename is specified
            (if filemode is unspecified, it defaults to 'a').
        format: Use the specified format string for the handler.
        datefmt: Use the specified date/time format.
        style: If a format string is specified, use this to specify the
            type of format string (possible values '%', '{', '$', for
            %-formatting, :meth:`str.format` and :class:`string.Template`
            - defaults to '%').
        level: Set the root logger level to the specified level.
        stream: Use the specified stream to initialize the StreamHandler. Note
            that this argument is incompatible with 'filename' - if both
            are present, 'stream' is ignored.
        handlers: If specified, this should be an iterable of already created
            handlers, which will be added to the root handler. Any handler
            in the list which does not have a formatter assigned will be
            assigned the formatter created in this function.

    Returns:
        None
    """
    if 'format' not in kwargs:
        kwargs['format'] = '[%(name)s] : %(asctime)s : %(levelname)s : %(message)s'
    if 'level' not in kwargs:
        kwargs['level'] = logging.INFO

    logging.basicConfig(**kwargs)


def get_filepath_recursive(dirpath, abspath=True, recursive=True):
    """获取目录下所有文件名 （默认递归）
        该函数主要用于需要一次性处理大量**相似文件**的情况

        该函数主要利用 `os.walk(path)` 实现，
        该函数会递归遍历 path 下的所有文件夹，并返回一个生成器

    Args:
        dirpath (str): 文件夹路径
        abspath (bool): 是否返回绝对路径
        recursive (bool): 是否递归

    Examples:
        >>> fs_gen = get_filepath_recursive('D:/Tmp')

    Returns:
        list
    """
    fs = []
    if recursive:
        for root, _, files in os.walk(dirpath):
            if abspath:
                fs.extend((os.path.join(root, file) for file in files))
            else:
                fs.extend(files)
    else:
        if abspath:
            fs.extend((os.path.join(dirpath, file) for file in os.listdir(dirpath)))
        else:
            fs.extend(os.listdir(dirpath))

    return fs


def maybe_download(url, to_path='D:/Tmp', filename=None, expected_byte=None):
    """下载文件到指定目录

    Args:
        url (str): 文件下载路径
        to_path (str): 下载到本地路径
        filename (str): 重命名文件
        expected_byte (int): 文件预期大小

    Returns:
        str: filepath

    Examples:

        >>> url = 'http://mattmahoney.net/dc/bbb.zip'
        >>> filepath = maybe_download(url, filename='b.zip')
        >>> fp = maybe_download(url, to_path='D:/Tmp/b', expected_byte=45370)

    """
    if filename is not None:
        filepath = os.path.join(maybe_mkdirs(to_path), filename)
    else:
        _, filename = os.path.split(url)
        filepath = os.path.join(maybe_mkdirs(to_path), filename)

    if not os.path.exists(filepath):
        urllib.request.urlretrieve(url, filepath)
        logging.info('File is downloading.')

        if expected_byte is not None:
            file_size = os.stat(filepath).st_size
            if file_size != expected_byte:
                logging.info('File has been damage, please download it manually.')
    else:
        logging.info('File is ready.')

    return filepath


def cycle_iter(iterator):
    """
    无限循环迭代器

    Args:
        iterator (Iterable): 可迭代对象

    Examples:

        >>> it = cycle_iter([1, 2, 3])
        >>> for _ in range(4):
        ...     print(next(it))
        1
        2
        3
        1

    """
    # while True:
    #     yield from iter(iterator)
    from itertools import cycle
    return cycle(iterator)


def system_is_windows():
    """
    If the system is windows, return True

    Examples:
        >>> if system_is_windows():
        ...     print("Windows")
        Windows
    """
    import platform
    return platform.system() == "Windows"


is_windows_system = system_is_windows()


def get_logger(name=None, fname=None, mode='a', level=logging.INFO, stream=None,
               fmt="[%(name)s] : %(asctime)s : %(levelname)s : %(message)s"):
    """创建一个 logger
        默认 log to console，如果同时指定了 fname，还会将日志输出到文件

    Examples:
        >>> logger = get_logger("Test", stream=sys.stdout, fmt="[%(name)s] : %(levelname)s : %(message)s")
        >>> logger.info("test")
        [Test] : INFO : test

    """
    logger = logging.Logger(name)
    logger.setLevel(level)

    fmt = logging.Formatter(fmt)

    ch = logging.StreamHandler(stream)
    ch.setFormatter(fmt)

    logger.addHandler(ch)

    if fname is not None:
        fh = logging.FileHandler(fname, mode)
        fh.setFormatter(fmt)

        logger.addHandler(fh)

    return logger


def to_unicode(txt, encoding='utf8', errors='strict'):
    """Convert text to unicode.

    Args:
        txt:
        encoding:
        errors:

    Returns:
        str

    """
    if sys.version_info[0] >= 3:
        unicode = str

    if isinstance(txt, unicode):
        return txt
    return unicode(txt, encoding, errors=errors)
